// -----------------------------------------------------------------------------
// Copyright 2009 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXPreferences.h"
#import "DGSMonXServer.h"


/// @brief This category declares private methods for the DGSMonXPreferences
/// class. 
@interface DGSMonXPreferences(Private)
/// @name Initialization and deallocation
//@{
- (void) dealloc;
//@}

/// @name NSWindow delegate
//@{
- (void)windowWillClose:(NSNotification*)notification;
//@}
@end


@implementation DGSMonXPreferences

// -----------------------------------------------------------------------------
/// @brief Initializes an DGSMonXPreferences object.
///
/// @note This is the designated initializer of DGSMonXPreferences.
// -----------------------------------------------------------------------------
- (id) init
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;

  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXPreferences object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  // These release() messages balance the retain() messages that were sent by
  // Cocoa's nib-loading code when it loaded MainMenu.xib.
  [m_preferencesDialog release];   // widgets in this dialog retain m_serversController
  [m_serversController release];   // retains the application model!

  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Registers application defaults in the "Registration" domain of
/// NSUserDefaults.
///
/// This class method must be called as early as possible, before any objects
/// of the application start accessing the user defaults.
///
/// @note dgsmonX does not support "resettable user defaults" in the preferences
/// dialog, therefore this method does not provide the NSUserDefaultsController
/// with initial values.
// -----------------------------------------------------------------------------
+ (void) setupDefaults
{
  // Make paranoid checks about the user defaults file to provide detailed
  // error handling. If any check fails, an exception with an appropriate
  // error message is thrown. The exception is caught and the error handler
  // is notified using the error message inside the exception.
  // Note: this mechanism also catches unexpected exceptions...
  @try
  {
    NSBundle* mainBundle = [NSBundle mainBundle];
    if (! mainBundle)
    {
      NSString* reason = @"Unable to access application bundle contents.";
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    NSString* defaultsPathName = [mainBundle pathForResource:registrationDomainDefaultsFileName ofType:nil];
    if (nil == defaultsPathName)
    {
      NSString* reason = [[[NSString stringWithString:@"The defaults file\n\n"]
        stringByAppendingString:registrationDomainDefaultsFileName]
        stringByAppendingString:@"\n\ncould not be found in the resources of the application bundle."];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    NSFileManager* fileManager = [NSFileManager defaultManager];
    BOOL isDirectory;  // use BOOL instead of bool due to method signature
    if (! [fileManager fileExistsAtPath:defaultsPathName isDirectory:&isDirectory])
    {
      NSString* reason = [[[NSString stringWithString:@"The defaults file\n\n"]
        stringByAppendingString:defaultsPathName]
        stringByAppendingString:@"\n\ndoes not exist."];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    if (! [fileManager isReadableFileAtPath:defaultsPathName])
    {
      NSString* reason = [[NSString stringWithString:@"No read access to defaults file\n\n"]
        stringByAppendingString:defaultsPathName];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    NSDictionary* defaultsDictionary = [NSDictionary dictionaryWithContentsOfFile:defaultsPathName];
    if (! defaultsDictionary)
    {
      NSString* reason = [[[NSString stringWithString:@"Defaults file\n\n"]
        stringByAppendingString:defaultsPathName]
        stringByAppendingString:@"\n\ndoes not contain an NSDictionary."];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    // Yippie, we did it...
    [[NSUserDefaults standardUserDefaults] registerDefaults:defaultsDictionary];
  }
  @catch(NSException* exception)
  {
    // Notify the error handler that an error has occurred
    [[NSNotificationCenter defaultCenter] postNotificationName:errorConditionOccurredNotification
                                                        object:[exception reason]];
  }
}

// -----------------------------------------------------------------------------
/// @brief Opens the user preferences dialog and displays it to the user.
///
/// @note The GUI controls in the dialog are "automatically" filled with values
/// by way of the "Cocoa Bindings" mechanism.
// -----------------------------------------------------------------------------
- (void) showPreferencesDialog
{
  // Make dgsmonX the active app so that the window we are going to display
  // is guaranteed to appear in front of other applications' windows.
  // Note: This is only necessary because dgsmonX is faceless and cannot become
  // the active application through normal user interaction.
  NSApplication* theApp = [NSApplication sharedApplication];
  [theApp activateIgnoringOtherApps:YES];

  // Prepare and run the dialog. The sheet is not attached to any window, it
  // is simply run as a modal window
  [theApp beginSheet:m_preferencesDialog modalForWindow:nil modalDelegate:nil didEndSelector:nil contextInfo:nil];
  [theApp runModalForWindow:m_preferencesDialog];

  // -------------------------------------------------
  // --- At this point the modal dialog is running ---
  // -------------------------------------------------

  // Clean up and close the dialog
  [theApp endSheet:m_preferencesDialog];
  // Remove dialog from screen
  [m_preferencesDialog orderOut:self];
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the user clicking the button labelled "Close" in the user
/// preferences dialog.
///
/// Simulates the user clicking the red "close" button, which leads to the
/// NSWindow delegate method windowWillClose:() being invoked.
// -----------------------------------------------------------------------------
- (void) preferencesDialogCloseClicked:(id)sender
{
  [m_preferencesDialog performClose:self];
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the user clicking the red "close" button of the NSWindow
/// that is the preferences dialog.
///
/// Closes the dialog.
///
/// @note The changes that the user made while the dialog was open have been
/// written to NSUserDefaults "automatically" by way of the "Cocoa Bindings"
/// mechanism.
///
/// @note This is an NSWindow delegate method.
// -----------------------------------------------------------------------------
- (void)windowWillClose:(NSNotification*)notification
{
  // Invoking this method closes the modal dialog and returns control to
  // showPreferencesDialog().
  [[NSApplication sharedApplication] stopModal];
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the user clicking the "test" button in the user
/// preferences dialog.
///
/// The currently selected server is tested if its URL can be found and if it
/// is possible to login with the current authenciation details.
// -----------------------------------------------------------------------------
- (void) testClicked:(id)sender
{
  if (! m_serversController)
    return;
  NSArray* selectedObjects = [m_serversController selectedObjects];
  if (! selectedObjects || [selectedObjects count] != 1)
    return;
  DGSMonXServer* selectedServer = (DGSMonXServer*)[selectedObjects objectAtIndex:0];
  if (! selectedServer)
    return;
  int connectionResult = [selectedServer testConnection];
  NSString* messageText;
  NSString* infoText;
  if (connectionResult >= 0)
  {
    messageText = @"Connection successful.";
    infoText = [NSString stringWithFormat:@"Number of running games found: %d", connectionResult];
  }
  else
  {
    messageText = @"Connection failed.";
    infoText = @"Reason: ";
    if (-1 == connectionResult)
      infoText = [infoText stringByAppendingString:@"URL not found"];
    else if (-2 == connectionResult)
      infoText = [infoText stringByAppendingString:@"URL found, but login not possible"];
    else
      infoText = [infoText stringByAppendingString:@"Unknown"];
  }
  NSAlert* alert = [NSAlert alertWithMessageText:messageText
                                   defaultButton:@"OK"
                                 alternateButton:nil
                                     otherButton:nil
                       informativeTextWithFormat:infoText];
  [alert setAlertStyle:NSInformationalAlertStyle];
  [alert runModal];
}

@end
