// -----------------------------------------------------------------------------
// Copyright 2009 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXServer.h"
#import "DGSMonXServerMenuItem.h"


/// @brief Counter used for handing out unique IDs to new DGSMonXServer
/// instances.
static int g_nextUniqueID = 0;


/// @brief This category declares private methods for the DGSMonXServer class.
@interface DGSMonXServer(Private)
/// @name Initialization and deallocation
//@{
- (void) dealloc;
//@}

/// @name Internal helpers
//@{
+ (int) newUniqueID;
- (void) resetResults;
//@}
@end


@implementation DGSMonXServer

// -----------------------------------------------------------------------------
/// @brief Initializes a DGSMonXServer object that is not being monitored, has
/// a default dummy name, uses the HTTP protocol and has none of the other
/// attributes set.
///
/// @note This initializer is required so that the NSArrayController in
/// MainMenu.xib is able to create new DGSMonXServer objects.
// -----------------------------------------------------------------------------
- (id) init
{
  // Invoke designated initializer, which will always return nil
  return [self initWithDictionary:nil];
}

// -----------------------------------------------------------------------------
/// @brief Initializes a DGSMonXServer object with User Defaults data stored
/// inside @a dictionary.
///
/// If @a dictionary is @e nil, the DGSMonXServer object is not being monitored,
/// has a dummy name, uses the HTTP protocol and has none of the other
/// attributes set.
///
/// Invoke the asDictionary() method to convert a DGSMonXServer object's User
/// Defaults attributes back into an NSDictionary suitable for storage in the
/// User Defaults system.
///
/// @note This is the designated initializer of DGSMonXServer.
// -----------------------------------------------------------------------------
- (id) initWithDictionary:(NSDictionary*)dictionary
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;
  else if (! dictionary)
  {
    m_monitored = false;
    m_name = newServerName;
    m_protocol = HttpProtocol;
    m_url = newServerURL;
    m_userID = nil;
    m_userPassword = nil;
  }
  else
  {
    // value has the type NSCFBoolean (try [value class]). It appears that this
    // can be treated as an NSNumber object, from which we can get the value
    // with boolValue().
    id value = [dictionary valueForKey:serverMonitoredKey];
    if (value)
      m_monitored = [value boolValue];
    else
      m_monitored = false;
    value = [dictionary valueForKey:serverNameKey];
    if (value)
      m_name = (NSString*)[value retain];
    else
      m_name = nil;
    // value has the type NSCFNumber (try [value class]). It appears that this
    // can be treated as an NSNumber object, from which we can get the value
    // with intValue().
    value = [dictionary valueForKey:serverProtocolKey];
    if (value)
      m_protocol = [value intValue];
    else
      m_protocol = HttpProtocol;   // use whatever is the first member of the DGSMonXServerProtocol enumeration
    value = [dictionary valueForKey:serverURLKey];
    if (value)
      m_url = (NSString*)[value retain];
    else
      m_url = nil;
    value = [dictionary valueForKey:userIDKey];
    if (value)
      m_userID = (NSString*)[value retain];
    else
      m_userID = nil;
    value = [dictionary valueForKey:userPasswordKey];
    if (value)
      m_userPassword = (NSString*)[value retain];
    else
      m_userPassword = nil;
  }

  // Initialize game state check results
  m_checkState = NonePerformedCheckState;
  m_checkSucceeded = false;
  m_errorDetails = nil;
  m_numberOfGames = -1;

  // Determine unique ID
  m_uniqueID = [DGSMonXServer newUniqueID];
  // Create the menu item, but let others add it to/remove it from the
  // appropriate menu
  m_menuItem = [[DGSMonXServerMenuItem alloc] initWithServer:self];

  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXServer object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  if (m_name)
    [m_name autorelease];
  if (m_url)
    [m_url autorelease];
  if (m_userID)
    [m_userID autorelease];
  if (m_userPassword)
    [m_userPassword autorelease];
  if (m_errorDetails)
    [m_errorDetails autorelease];
  [m_menuItem autorelease];
  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Hands out a new unique ID.
// -----------------------------------------------------------------------------
+ (int) newUniqueID
{
  return g_nextUniqueID++;
}

// -----------------------------------------------------------------------------
/// @brief Returns the unique ID of this DGSMonXServer.
// -----------------------------------------------------------------------------
- (int) uniqueID
{
  return m_uniqueID;
}

// -----------------------------------------------------------------------------
/// @brief Returns whether or not this DGSMonXServer is being monitored.
// -----------------------------------------------------------------------------
- (bool) monitored
{
  return m_monitored;
}

// -----------------------------------------------------------------------------
/// @brief Sets whether this DGSMonXServer is being monitored to @a monitored.
// -----------------------------------------------------------------------------
- (void) setMonitored:(bool)monitored
{
  m_monitored = monitored;
}

// -----------------------------------------------------------------------------
/// @brief Returns the name of this DGSMonXServer. The name is a string that
/// should be used to represent this DGSMonXServer in the GUI.
// -----------------------------------------------------------------------------
- (NSString*) name
{
  if (m_name)
    return [[m_name retain] autorelease];
  else
    return nil;
}

// -----------------------------------------------------------------------------
/// @brief Sets the name of this DGSMonXServer to @a name.
// -----------------------------------------------------------------------------
- (void) setName:(NSString*)name
{
  if (m_name == name)
    return;
  if (m_name)
    [m_name autorelease];
  if (name)
    m_name = [name retain];
  else
    m_name = nil;
}

// -----------------------------------------------------------------------------
/// @brief Returns the protocol that this DGSMonXServer currently uses.
///
/// The return value of this method is a member of the enumeration
/// #DGSMonXServerProtocol.
// -----------------------------------------------------------------------------
- (int) protocol
{
  return m_protocol;
}

// -----------------------------------------------------------------------------
/// @brief Changes the protocol that this DGSMonXServer currently uses to
/// to @a protocol.
///
/// @a protocol must be a member of the enumeration #DGSMonXServerProtocol.
// -----------------------------------------------------------------------------
- (void) setProtocol:(int)protocol
{
  m_protocol = protocol;
}

// -----------------------------------------------------------------------------
/// @brief Returns the URL of this DGSMonXServer.
// -----------------------------------------------------------------------------
- (NSString*) url
{
  if (m_url)
    return [[m_url retain] autorelease];
  else
    return nil;
}

// -----------------------------------------------------------------------------
/// @brief Sets the URL of this DGSMonXServer to @a url.
// -----------------------------------------------------------------------------
- (void) setUrl:(NSString*)url
{
  if (m_url == url)
    return;
  if (m_url)
    [m_url autorelease];
  if (url)
    m_url = [url retain];
  else
    m_url = nil;
}

// -----------------------------------------------------------------------------
/// @brief Returns the user ID of this DGSMonXServer.
// -----------------------------------------------------------------------------
- (NSString*) userID
{
  if (m_userID)
    return [[m_userID retain] autorelease];
  else
    return nil;
}

// -----------------------------------------------------------------------------
/// @brief Sets the user ID of this DGSMonXServer to @a userID.
// -----------------------------------------------------------------------------
- (void) setUserID:(NSString*)userID
{
  if (m_userID == userID)
    return;
  if (m_userID)
    [m_userID autorelease];
  if (userID)
    m_userID = [userID retain];
  else
    m_userID = nil;
}

// -----------------------------------------------------------------------------
/// @brief Returns the user password of this DGSMonXServer.
// -----------------------------------------------------------------------------
- (NSString*) userPassword
{
  if (m_userPassword)
    return [[m_userPassword retain] autorelease];
  else
    return nil;
}

// -----------------------------------------------------------------------------
/// @brief Sets the user password of this DGSMonXServer to @a userPassword.
// -----------------------------------------------------------------------------
- (void) setUserPassword:(NSString*)userPassword
{
  if (m_userPassword == userPassword)
    return;
  if (m_userPassword)
    [m_userPassword autorelease];
  if (userPassword)
    m_userPassword = [userPassword retain];
  else
    m_userPassword = nil;
}

// -----------------------------------------------------------------------------
/// @brief Returns this DGSMonXServer object's User Defaults attributes as an
/// unretained dictionary suitable for storage in the User Defaults system.
// -----------------------------------------------------------------------------
- (NSDictionary*) asDictionary
{
  // Set up unretained empty dictionary
  NSMutableDictionary* dictionary = [NSMutableDictionary dictionary];
  // setValue:forKey:() allows for nil values, so we use that instead of
  // setObject:forKey:() which is less forgiving and would force us to check
  // for nil values.
  // Note: Use NSNumber to represent int and bool values as an object.
  [dictionary setValue:[NSNumber numberWithBool:[self monitored]] forKey:serverMonitoredKey];
  [dictionary setValue:[self name] forKey:serverNameKey];
  [dictionary setValue:[NSNumber numberWithInt:[self protocol]] forKey:serverProtocolKey];
  [dictionary setValue:[self url] forKey:serverURLKey];
  [dictionary setValue:[self userID] forKey:userIDKey];
  [dictionary setValue:[self userPassword] forKey:userPasswordKey];
  return [[dictionary retain] autorelease];
}

// -----------------------------------------------------------------------------
/// @brief Returns the state of this DGSMonXServer, in relation to game state
/// checks being performed on this server.
///
/// The return value of this method is a member of the enumeration
/// #DGSMonXServerCheckState.
///
/// Useful game state check results are available only if the return value is
/// #DidFinishCheckState.
// -----------------------------------------------------------------------------
- (int) checkState;
{
  return m_checkState;
}

// -----------------------------------------------------------------------------
/// @brief Sets the state of this DGSMonXServer, in relation to game state
/// checks being performed on this server, to @a checkState.
///
/// Also resets all game state check results to default values.
///
/// @a checkState must be a member of the enumeration #DGSMonXServerCheckState.
///
/// @note This method is intended to be invoked by DGSMonXServerCheck only.
// -----------------------------------------------------------------------------
- (void) setCheckState:(int)checkState
{
  m_checkState = checkState;
  [self resetResults];
}

// -----------------------------------------------------------------------------
/// @brief Returns whether or not the most recent game state check succeeded.
///
/// Returns false if no game state check has been performed yet, or while one
/// is still in progress.
// -----------------------------------------------------------------------------
- (bool) checkSucceeded;
{
  return m_checkSucceeded;
}

// -----------------------------------------------------------------------------
/// @brief Changes whether or not the most recent game state check succeeded to
/// @a checkSucceedd.
///
/// @note This method is intended to be invoked by DGSMonXServerCheck only.
// -----------------------------------------------------------------------------
- (void) setCheckSucceeded:(bool)checkSucceeded;
{
  m_checkSucceeded = checkSucceeded;
}

// -----------------------------------------------------------------------------
/// @brief Returns a string with details about why the most recent game state
/// check failed.
///
/// Returns nil if no game state check has been performed yet, or while one is
/// still in progress. Also returns nil if the most recent game state check
/// succeeded.
// -----------------------------------------------------------------------------
- (NSString*) errorDetails;
{
  if (m_errorDetails)
    return [[m_errorDetails retain] autorelease];
  else
    return nil;
}

// -----------------------------------------------------------------------------
/// @brief Sets details about why the most recent game state check failed to
/// @a errorDetails.
///
/// @note This method is intended to be invoked by DGSMonXServerCheck only.
// -----------------------------------------------------------------------------
- (void) setErrorDetails:(NSString*)errorDetails
{
  if (m_errorDetails == errorDetails)
    return;
  if (m_errorDetails)
    [m_errorDetails autorelease];
  if (errorDetails)
    m_errorDetails = [errorDetails retain];
  else
    m_errorDetails = nil;
}

// -----------------------------------------------------------------------------
/// @brief Returns the number of games that were found during the most recent
/// game state check for which it is the user's turn.
///
/// Returns -1 if no game state check has been performed yet, or while one is
/// still in progress. Also returns -1 if the most recent game state check
/// failed.
// -----------------------------------------------------------------------------
- (int) numberOfGames;
{
  return m_numberOfGames;
}

// -----------------------------------------------------------------------------
/// @brief Changes the number of games that were found during the most recent
/// game state check for which it is the user's turn, to @a numberOfGames.
///
/// @note This method is intended to be invoked by DGSMonXServerCheck only.
// -----------------------------------------------------------------------------
- (void) setNumberOfGames:(int)numberOfGames
{
  m_numberOfGames = numberOfGames;
}

// -----------------------------------------------------------------------------
/// @brief Returns true if numberOfGames() returns a value of 1 or greater.
///
/// Returns false if no game state check has been performed yet, or while one
/// is still in progress. Also returns false if the most recent game state
/// check failed.
///
/// @note This is a convenience method, the value returned does not have its own
/// key (in a key-value coding context).
// -----------------------------------------------------------------------------
- (bool) itsMyTurn;
{
  return ([self numberOfGames] > 0);
}

// -----------------------------------------------------------------------------
/// @brief Resets all game state check results to initial values.
///
/// This method is intended to be invoked after the state of this DGSMonXServer,
/// in relation to game state checks being performed on this server, is
/// changed (e.g. by setCheckState:()). This method does @b not change the
/// server state itself!
///
/// This method invokes set... accessor methods so that the key-value coding
/// aware "things" may pick up the changes.
///
/// @note This is an internal helper.
// -----------------------------------------------------------------------------
- (void) resetResults
{
  [self setCheckSucceeded:false];
  [self setErrorDetails:nil];
  [self setNumberOfGames:-1];
}

// -----------------------------------------------------------------------------
/// @brief Returns the menu item used to represent this DGSMonXServer.
// -----------------------------------------------------------------------------
- (DGSMonXServerMenuItem*) menuItem
{
  if (m_menuItem)
    return [[m_menuItem retain] autorelease];
  else
    return nil;
}

@end
