// -----------------------------------------------------------------------------
// Copyright 2009 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXServerCheck.h"
#import "DGSMonXServer.h"


/// @brief This category declares private methods for the DGSMonXServerCheck
/// class.
@interface DGSMonXServerCheck(Private)
/// @name Initialization and deallocation
//@{
- (void) dealloc;
//@}

/// @name Game state check methods
//@{
- (void) check;
- (void) checkFails:(NSString*)errorDetails;
- (void) checkSucceeds:(int)numberOfGames;
//@}

/// @name Adopts the informal protocol NSURLConnectionDelegate
//@{
- (void) connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response;
- (void) connection:(NSURLConnection*)connection didReceiveData:(NSData*)data;
- (void) connection:(NSURLConnection*)connection didFailWithError:(NSError*)error;
- (void) connectionDidFinishLoading:(NSURLConnection*)connection;
//@}

/// @name Cookie handling
//@{
+ (NSArray*) removeCookies:(NSURL*)url;
- (void) restoreCookieState;
//@}

/// @name Internal helpers
//@{
- (NSURLRequest*) request:(DGSMonXServer*)server forPage:(NSString*)pageName;
- (void) logURLDetails:(NSURL*)url;
- (void) cookieCleanup;
- (void) callbackDelegate;
//@}
@end


@implementation DGSMonXServerCheck

// -----------------------------------------------------------------------------
/// @brief Initializes a DGSMonXServerCheck object that performs a game state
/// check on @a server and calls back @a delegate when the check is done.
///
/// @a server and @a delegate must not be @e nil.
///
/// @note This is the designated initializer of DGSMonXServerCheck.
// -----------------------------------------------------------------------------
- (id) initWithServer:(DGSMonXServer*)server delegate:(id)delegate
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;

  // Don't retain to prevent a retain cycle
  m_server = server;
  // Don't retain. The delegate mechanism as it is used by Cocoa advocates
  // embedding the delegate as a "weak", i.e. unretained, reference.
  m_delegate = delegate;

  // This is assigned when the game state check begins
  m_connection = nil;
  // This is assigned/retained/released during a HTTP protocol exchange
  m_receivedResponse = nil;
  // Reusable container that will hold data received during a HTTP protocol
  // exchange. The container is initially created empty.
  m_receivedData = [[NSMutableData data] retain];
  // Initialized by check:(), evaluated and released by restoreCookieState()
  m_cookiesToRestore = nil;

  // Launch game state check. The method returns immediately and the check
  // proceeds asynchronously, due to the asynchronous nature of NSURLConnection.
  [self check];

  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXServer object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  [m_receivedData autorelease];
  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Returns the server for which this DGSMonXServerCheck performs, or
/// has performed, a game state check.
// -----------------------------------------------------------------------------
- (DGSMonXServer*) server
{
  return m_server;
}

// -----------------------------------------------------------------------------
/// @brief Launches the game state check.
///
/// The details about the checking procedure are part of the class
/// documentation.
// -----------------------------------------------------------------------------
- (void) check
{
  // Mark the game state check as being in progress. From now on, only the
  // checkFails:() and checkSucceeds:() methods should be used to mark the check
  // as finished (and to callback the delegate).
  [m_server setCheckState:InProgressCheckState];

  // Construct request for login page
  NSString* loginPageWithQueryString = [NSString stringWithFormat:@"%@?userid=%@&passwd=%@",
                                        loginPage,
                                        [m_server userID],
                                        [m_server userPassword]];
  NSURLRequest* loginPageRequest = [self request:m_server forPage:loginPageWithQueryString];
  if (! loginPageRequest)
  {
    [self checkFails:[NSString stringWithFormat:@"Failed to construct HTTP request for login page"]];
    return;
  }

  // Temporarily remove cookies that exist at the moment for the DGS server.
  // They will be restored when the game state check finishes.
  m_cookiesToRestore = [DGSMonXServerCheck removeCookies:[loginPageRequest URL]];
  if (m_cookiesToRestore)
    [m_cookiesToRestore retain];

  // Create the connection that retrieves the login page. The connection
  // immediately starts loading the data.
  m_connection = [[NSURLConnection alloc] initWithRequest:loginPageRequest delegate:self];
  if (! m_connection)
  {
    [self checkFails:[NSString stringWithFormat:@"Connection for login page could not be initialized"]];
    return;
  }
}

// -----------------------------------------------------------------------------
/// @brief Terminates the game state check and calls back the delegate. The
/// check is considered to have failed.
///
/// @a errorDetails describes the reason for the failure.
// -----------------------------------------------------------------------------
- (void) checkFails:(NSString*)errorDetails
{
  [m_server setCheckState:DidFinishCheckState];
  [m_server setCheckSucceeded:false];
  [m_server setErrorDetails:errorDetails];

  [self cookieCleanup];
  [self callbackDelegate];
}

// -----------------------------------------------------------------------------
/// @brief Terminates the game state check and calls back the delegate. The
/// check is considered to have succeeded.
///
/// @a numberOfGames is the number of games detected during the check for which
/// it is the user's turn.
// -----------------------------------------------------------------------------
- (void) checkSucceeds:(int)numberOfGames
{
  [m_server setCheckState:DidFinishCheckState];
  [m_server setCheckSucceeded:true];
  [m_server setNumberOfGames:numberOfGames];

  [self cookieCleanup];
  [self callbackDelegate];
}

// -----------------------------------------------------------------------------
/// @brief Terminates the game state check and calls back the delegate. The
/// check is considered to have been cancelled.
///
/// Calling this method has no effect if no game state check is in progress.
// -----------------------------------------------------------------------------
- (void) cancel
{
  if ([m_server checkState] != InProgressCheckState)
    return;

  [m_connection cancel];   // we should not get any more delegate messages

  [m_server setCheckState:WasCancelledCheckState];

  [self cookieCleanup];
  [self callbackDelegate];
}

// -----------------------------------------------------------------------------
/// @brief Performs cookie cleanup operations after a game state check has
/// finished.
///
/// @note This is an internal helper.
// -----------------------------------------------------------------------------
- (void) cookieCleanup
{
  [self restoreCookieState];
  // Release after cleanup
  if (m_connection)
  {
    [m_connection autorelease];
    m_connection = nil;
  }
  if (m_receivedResponse)
  {
    [m_receivedResponse autorelease];
    m_receivedResponse = nil;
  }
  if (m_cookiesToRestore)
  {
    [m_cookiesToRestore autorelease];
    m_cookiesToRestore = nil;
  }
}

// -----------------------------------------------------------------------------
/// @brief Sends the callback message to the delegate after a game state check
/// has finished.
///
/// @note This is an internal helper.
// -----------------------------------------------------------------------------
- (void) callbackDelegate
{
  [m_delegate gameStateCheckDidFinish:self];
  m_delegate = nil;
}

// -----------------------------------------------------------------------------
/// @brief Sent when @a connection has received sufficient data from the server
/// to construct @a response.
///
/// This implementation does the following things:
/// - initialize a data structure that will later be used by
///   connection:didReceiveData:() to store content data that is incrementally
///   received from the server
/// - store @a response so that connectionDidFinishLoading:() can evaluate
///   which page the data came from
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response
{
  if (m_receivedResponse)
    [m_receivedResponse autorelease];
  m_receivedResponse = [response retain];
  [m_receivedData setLength:0];
}

// -----------------------------------------------------------------------------
/// @brief Sent as @a connection loads data incrementally.
///
/// This implementation appends @a data to the data structure previously
/// initialized by connection:didReceiveResponse:().
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveData:(NSData*)data
{
  [m_receivedData appendData:data];
}

// -----------------------------------------------------------------------------
/// @brief Sent when @a connection fails to load its request successfully.
///
/// This implementaton terminates the game state check by invoking
/// checkFails:(). The error details are taken from the message encapsulated by
/// @a error.
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didFailWithError:(NSError*)error
{
  // This happens for instance if there is no network connection to the server
  // (e.g. no internet connection, or the server name is wrong)
  [self checkFails:[error localizedDescription]];
}

// -----------------------------------------------------------------------------
/// @brief Sent when @a connection has finished loading successfully.
///
/// This implementaton checks the NSURLResponse object that was cached by
/// connection:didReceiveResponse:() in order to determine what should be done
/// next. The NSURLResponse may refer to the following pages:
/// - status.php: The data collected by connection:didReceiveData:() is
///   evaluated
///   - If evaluation succeeds, the game state check is terminated by invoking
///     checkSucceeds:()
///   - If evaluation fails, the game state check is terminated by invoking
///     checkFails:()
/// - error.php: It is assumed that login failed. The game state check is
///   terminated by invoking checkFails:(). The error details are taken from
///   the query string that has been supplied to error.php.
/// - Anything else: It is assumed that login succeeded, but that for some
///   reason the default page after login is not status.php. Another
///   NSURLConnection cycle is triggered in order to retrieve status.php.
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connectionDidFinishLoading:(NSURLConnection*)connection
{
  // Check fails if there is an HTTP error status code
  int statusCode = [(NSHTTPURLResponse*)m_receivedResponse statusCode];
  if (statusCode >= 400 && statusCode <= 599)
  {
    [self checkFails:[NSString stringWithFormat:@"HTTP status code %d", statusCode]];
    return;
  }

  // Extract the name of the page for which we have data
  // Note: We cannot use [NSString pathComponents] because that method works on
  // file system paths, not URL paths.
  NSURL* responseURL = [m_receivedResponse URL];
  NSString* pageName = (NSString*)[[[responseURL path] componentsSeparatedByString:@"/"] lastObject];

  if (NSOrderedSame == [pageName compare:statusPage])
  {
    // Convert data to string
    /// @todo Find a function that converts encoding names (NSString*) into
    /// encoding enumeration values for us.
    NSString* encodingName = [m_receivedResponse textEncodingName];
    NSStringEncoding encoding;
    if (! encodingName)
      encoding = NSISOLatin1StringEncoding;
    else if (NSOrderedSame == [encodingName compare:@"utf-8"])
      encoding = NSUTF8StringEncoding;
    else if (NSOrderedSame == [encodingName compare:@"iso-8859-1"])
      encoding = NSISOLatin1StringEncoding;
    else
      encoding = NSISOLatin1StringEncoding;
    NSString* dataAsString = [[NSString alloc] initWithData:m_receivedData encoding:encoding];
    if (! dataAsString)
    {
      [self checkFails:[NSString stringWithFormat:@"Unable to decode status page data using encoding %@", encodingName]];
      return;
    }

    /// @todo Make parsing of the status page more robust. The current approach
    /// of searching inside a string is rather fragile because it does not take
    /// the structure of the HTML document into account. Thus, if the string
    /// "\"sectGames\"" appears somewhere in the document where we don't expect
    /// it (e.g. in a comment), we might get wrong results. The best approach
    /// would be to use a DOM parser to parse the status page. The structure
    /// of the status page at the time of writing this looked something like
    /// this (note that the class attribute is probably used to alternate table
    /// rows for coloring):
    /// @code
    ///   <table id="gameTable" class="Table">
    ///     <tbody>
    ///       <tr id="TableHead" class="Head">
    ///       <tr class="Row1" ondblclick="javascript: ............ ">
    ///       <tr class="Row2" ondblclick="javascript: ............ ">
    ///     </tbody
    ///   </table>
    /// @endcode

    /// @todo Test if the following compiles with the 10.5 SDK. It *should*
    /// compile, because the API docs say for both types NSUInteger and
    /// NSStringCompareOptions: "Available in Mac OS X v10.5 and later".
#if (MAC_OS_X_VERSION_MAX_ALLOWED <= MAC_OS_X_VERSION_10_4)
    unsigned int dataLength;
    unsigned compareOptions;
#else
    NSUInteger dataLength;
    NSStringCompareOptions compareOptions;
#endif
    dataLength = [dataAsString length];
    compareOptions = NSCaseInsensitiveSearch;
    NSRange searchRange = {0, dataLength};

    // Look for the section that lists games for which it is the user's turn.
    // We expect this section to always appear, even if no games are found.
    NSRange resultRange = [dataAsString rangeOfString:@"\"sectGames\"" options:compareOptions range:searchRange];
    if (resultRange.location == NSNotFound)
    {
      [self checkFails:@"Unable to parse status page"];
      return;
    }

    // From the position where the game list section starts, look for
    // occurrences where sgf.php is invoked with a query string. We assume
    // that each of these occurrences marks one game for which it is the
    // user's turn, therefore by counting the occurrences of sgf.php we
    // can determine the total number of games for which it is the user's
    // turn.
    int numberOfGames = 0;
    while (true)
    {
      // Start searching from the position where the last search ended
      searchRange.location = resultRange.location + resultRange.length;
      searchRange.length = dataLength - searchRange.location;
      resultRange = [dataAsString rangeOfString:@"\"sgf.php?" options:compareOptions range:searchRange];
      if (resultRange.location == NSNotFound)
        break;
      ++numberOfGames;
    }
    [self checkSucceeds:numberOfGames];
  }
  else if (NSOrderedSame == [pageName compare:errorPage])
  {
    /// @todo Improve parsing; I have seen errors such as the following:
    /// "error.php?err=receiver_not_found#preview". Also, it is unclear whether
    /// a second parameter is possible, such as this:
    /// "error.php?err=foo&parameter2=bar"
    NSString* error = (NSString*)[[pageName componentsSeparatedByString:@"?err="] lastObject];
    [self checkFails:error];
  }
  else
  {
    // Create another connection that retrieves the status page. The connection
    // immediately starts loading the data.
    // Note: This part exists to guard against future changes in the behaviour
    // of the DGS software (or behaviour that I did not know about), i.e. when
    // the login page does not automatically redirect to the status page.
    NSURLRequest* request = [self request:m_server forPage:statusPage];
    if (! request)
      [self checkFails:@"Failed to construct HTTP request for status page"];
    else
    {
      [m_connection autorelease];
      m_connection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
      if (! m_connection)
        [self checkFails:@"Connection for status page could not be initialized"];
    }
  }
}

// -----------------------------------------------------------------------------
/// @brief Removes cookies that are recognized as DGS cookies from the shared
/// cookie storage system.
///
/// @a url is an URL that points to the DGS server that we are performing the
/// game state check on. It is used as an initial filter to get only cookies
/// that are valid for this DGS server.
///
/// @retval An array with cookies that have been removed
/// @retval nil if no cookies have been removed
///
/// @note This is an internal helper used for cookie handling.
///
/// @see The chapter "Game state check procedure" in the class documentation has
/// a detailed discussion about the hows and whys of cookie handling.
// -----------------------------------------------------------------------------
+ (NSArray*) removeCookies:(NSURL*)url
{
  NSMutableArray* removedCookies = nil;

  NSHTTPCookieStorage* storage = [NSHTTPCookieStorage sharedHTTPCookieStorage];
  NSArray* cookies = [storage cookiesForURL:url];
  NSEnumerator* enumerator = [cookies objectEnumerator];
  while (enumerator)
  {
    NSHTTPCookie* cookie = (NSHTTPCookie*)[enumerator nextObject];
    if (! cookie)
      break;
    // We look for two cookies that we know about by name
    if (NSOrderedSame != [[cookie name] compare:handleCookie] &&
        NSOrderedSame != [[cookie name] compare:sessionCodeCookie])
      continue;
    if (! removedCookies)
      removedCookies = [NSMutableArray array];
    // NSMutableArray retains (and releases) the cookie object for us
    [removedCookies addObject:cookie];
    [storage deleteCookie:cookie];
  }

  if (removedCookies)
    return [[removedCookies retain] autorelease];
  else
    return nil;
}

// -----------------------------------------------------------------------------
/// @brief Restores the shared cookie storage system to the state it was in
/// before the game state check started.
///
/// The following restore operations are performed:
/// - Remove any cookies that have been generated by the game state check. The
///   shared cookie storage system is examined if it contains any cookies for
///   the URL of the response that was processed last during the game state
///   check. Any cookies found are removed. This approach should be OK if we
///   assume that DGS cookies apply to all pages of the DGS server alike - the
///   URL of the last response is therefore as good as any other.
/// - Restores DGS cookies that existed for the DGS server before the game
///   state check was launched.
///
/// @note This is an internal helper used for cookie handling.
///
/// @see The chapter "Game state check procedure" in the class documentation has
/// a detailed discussion about the hows and whys of cookie handling.
// -----------------------------------------------------------------------------
- (void) restoreCookieState
{
  // Remove DGS cookies that have been generated by the game state check
  if (m_receivedResponse)
  {
    NSURL* lastResponseURL = [m_receivedResponse URL];
    [DGSMonXServerCheck removeCookies:lastResponseURL];
  }

  // Restore DGS cookies that were present before the game state check was
  // launched
  if (m_cookiesToRestore)
  {
    NSHTTPCookieStorage* storage = [NSHTTPCookieStorage sharedHTTPCookieStorage];
    NSEnumerator* enumerator = [m_cookiesToRestore objectEnumerator];
    while (enumerator)
    {
      NSHTTPCookie* cookie = (NSHTTPCookie*)[enumerator nextObject];
      if (! cookie)
        break;
      [storage setCookie:cookie];
    }
  }
}

// -----------------------------------------------------------------------------
/// @brief Returns an NSURLRequest object that refers to the page named
/// @a pageName located on the Dragon Go Server @a server.
///
/// Returns @e nil if any error occurs.
///
/// @note This is an internal helper.
// -----------------------------------------------------------------------------
- (NSURLRequest*) request:(DGSMonXServer*)server forPage:(NSString*)pageName
{
  // Construct the URL that is the base for all pages of the Dragon Go Server
  NSString* baseURL = nil;
  int protocol = [server protocol];
  switch (protocol)
  {
    case HttpProtocol:
      baseURL = @"http://";
      break;
    case HttpsProtocol:
      baseURL = @"https://";
      break;
    default:
      return nil;
  }
  baseURL = [baseURL stringByAppendingString:[server url]];

  // Construct the request
  NSString* pageURL = [baseURL stringByAppendingString:pageName];
  NSURLRequest* request = [NSURLRequest requestWithURL:[NSURL URLWithString:pageURL]];
  return [[request retain] autorelease];
}

// -----------------------------------------------------------------------------
/// @brief Prints all details of @a url to NSLog.
///
/// @note This is an internal helper used for debugging purposes.
// -----------------------------------------------------------------------------
- (void) logURLDetails:(NSURL*)url
{
  NSStringEncoding encoding;
  NSLog(@"absoluteString = %@", [url absoluteString]);
  NSLog(@"absoluteURL = %@", [NSString stringWithContentsOfURL:[url absoluteURL] usedEncoding:&encoding error:NULL]);
  NSLog(@"baseURL = %@", [NSString stringWithContentsOfURL:[url baseURL] usedEncoding:&encoding error:NULL]);
  NSLog(@"fragment = %@", [url fragment]);
  NSLog(@"host = %@", [url host]);
  NSLog(@"parameterString = %@", [url parameterString]);
  NSLog(@"password = %@", [url password]);
  NSLog(@"path = %@", [url path]);
  NSLog(@"port = %@", [url port]);
  NSLog(@"query = %@", [url query]);
  NSLog(@"relativePath = %@", [url relativePath]);
  NSLog(@"relativeString = %@", [url relativeString]);
  NSLog(@"resourceSpecifier = %@", [url resourceSpecifier]);
  NSLog(@"scheme = %@", [url scheme]);
  NSLog(@"standardizedURL = %@", [NSString stringWithContentsOfURL:[url standardizedURL] usedEncoding:&encoding error:NULL]);
  NSLog(@"user = %@", [url user]);
}

@end
